import copy
import re

import numpy as np
from rdkit.Chem import GetAdjacencyMatrix, AddHs, RemoveHs
from rdkit.Chem.rdchem import ChiralType

from glyles.glycans.factory.factory_o import OpenFactory, c1_finder
from glyles.glycans.utils import find_longest_c_chain, opposite_chirality
from glyles.grammar.GlycanLexer import GlycanLexer

if not hasattr(GlycanLexer, "MOD"):
    GlycanLexer.MOD = GlycanLexer.QMARK + 1


def get_indices(names, types):
    sac_index = types.index(GlycanLexer.SAC)

    # if there's nothing to do, return
    if len(types) == sac_index + 1 or (len(types) > sac_index + 1 and types[sac_index + 1] != GlycanLexer.SAC):
        return sac_index, None

    # identify the root and the new size of the monomer by identifying the indices of their descriptions
    if names[sac_index + 1] in ["Pen", "Hex", "Hep", "Oct"]:
        return sac_index, sac_index + 1


def check_for_resizing(monomer, names, types):
    """
    Change the monosaccharide in case it's indicated that the monosaccharide has more carbon atoms than normal.

    Args:
        monomer (Monomer): Monomer to be modified according to functional groups
        names (List[str]): List of the names of the functional groups
        types (List[int]): List of type definitions from the recipe of a monosaccharide

    Returns:
        Nothing
    """
    sac_index, len_index = get_indices(names, types)

    # extract the orientations of the hydroxy-groups along the enlarged chain
    if sac_index != 0 and names[sac_index - 1].count("L") + \
            names[sac_index - 1].count("D") == len(names[sac_index - 1]):
        orientations = names[sac_index - 1]
    else:
        orientations = ""

    # determine the chain that has to be added to the last carbon atom
    c_count = np.count_nonzero(monomer.x[:, 0] == 6)
    if names[len_index] == "Pen":
        count = 5
    elif names[len_index] == "Hex":
        count = 6
    elif names[len_index] == "Hep":
        count = 7
    elif names[len_index] == "Oct":
        count = 8
    else:
        count = c_count
    extension = "[C?H](O)" * (count - c_count) + "CO"

    # insert the orientations
    for c in orientations[:-1]:
        if c == "L":
            extension = extension.replace("[C?H]", "[C@@H]", 1)
        if c == "D":
            extension = extension.replace("[C?H]", "[C@H]", 1)
    extension = extension.replace("[C?H]", "C")

    # find oxygen of c6 to delete it and find c6 to replace it
    # TODO: Check if this is actually valid for all cases
    c_id = np.where(monomer.x[:, 1] == np.max(monomer.x[monomer.x[:, 0] == 6, 1]).item())[0].item()
    ox_id = monomer.find_oxygen(position=c_id)

    # if the carbon to be extended is part of the ring, put the extension into a side_chain
    c_has_ox = ox_id != c_id
    c_in_ring = bool(monomer.x[c_id, 2] & 0b1)

    if c_in_ring and c_has_ox:  # no example known (yet)
        smiles = monomer.to_smiles(0, root_idx=100)
    elif c_in_ring and not c_has_ox:  # AraHex
        tmp = AddHs(monomer.structure)
        for n in tmp.GetAtomWithIdx(c_id).GetNeighbors():
            if n.GetAtomicNum() == 1:
                n.SetAtomicNum(50)
                break
        monomer.structure = RemoveHs(tmp)
        extension = extension[extension.index(")") + 1:]
        carb_atom = monomer.structure.GetAtomWithIdx(c_id)
        prev_c_id = int(np.where(monomer.x[:, 1] == (max(monomer.x[monomer.x[:, 0] == 6, 1]) - 1))[0])
        if carb_atom.GetChiralTag() == ChiralType.CHI_UNSPECIFIED:
            # carb_atom.SetChiralTag(opposite_chirality(monomer.structure.GetAtomWithIdx(prev_c_id).GetChiralTag()))
            carb_atom.SetChiralTag(monomer.structure.GetAtomWithIdx(prev_c_id).GetChiralTag())
        smiles = monomer.to_smiles(0, root_idx=100)
        smiles = re.sub(r"\[SnH*\d*\]", extension, smiles)
    elif not c_in_ring and c_has_ox:  # LDManHep/DDManHep
        monomer.structure.GetAtomWithIdx(c_id).SetAtomicNum(50)
        monomer.structure.GetAtomWithIdx(ox_id).SetAtomicNum(1)
        smiles = monomer.to_smiles(0, root_idx=100)
        smiles = re.sub(r"\[SnH*\d*\]", extension, smiles)
    elif not c_in_ring and not c_has_ox:  # no example known (yet)
        monomer.structure.GetAtomWithIdx(c_id).SetAtomicNum(50)
        smiles = monomer.to_smiles(0, root_idx=100)
        smiles = re.sub(r"\[SnH*\d*\]", extension, smiles)
    else:
        smiles = monomer.to_smiles(0, root_idx=100)

    # update the monomer
    monomer.smiles = smiles
    monomer.structure = None
    monomer.get_structure()


def check_for_open_form(monomer, names, types, longer=False):
    """
    This method checks if the monomer is converted into an open form.

    Args:
        monomer (Monomer): Monomer to be modified according to functional groups
        names (List[str]): List of the names of the functional groups
        types (List[int]): List of type definitions from the recipe of a monosaccharide
        longer (bool): True, if the glycan is longer/shorted than normal

    Returns:
        Nothing
    """
    # identify the descriptions of what to do and determine the new SMILES string
    sac_index, len_index = get_indices(names, types)
    params = copy.copy(OpenFactory()[names[sac_index] + "-ol"])

    if longer:
        last = params["smiles"].find("C") + 1
        a_type = np.array([a.GetAtomicNum() for a in monomer.structure.GetAtoms()])
        adjacency = GetAdjacencyMatrix(monomer.structure, useBO=True)
        c_atoms = np.where(a_type == 6)[0].tolist()
        count = len(find_longest_c_chain(c_atoms, adjacency, a_type))
        if names[len_index] == "Pen":
            maximum = 5
        elif names[len_index] == "Hex":
            maximum = 6
        elif names[len_index] == "Hep":
            maximum = 7
        elif names[len_index] == "Oct":
            maximum = 8
        else:
            maximum = count
        params["smiles"] = params["smiles"][:last] + "C(O)" * max((maximum - count), 0) + params["smiles"][last:]

    if "-onic" in names or "-aric" in names:
        params["smiles"] = params["smiles"].replace("C", "C(=O)", 1)
    if ("-aric" in names or "-ulosaric" in names) and names[sac_index] != "Qui":
        params["smiles"] = params["smiles"][:-1] + "(=O)O"
    if "-ulosonic" in names or "-ulosaric" in names:
        params["smiles"] = re.sub(r"OC\[*C@*[H\]]*\(O\)", "OC(=O)C(=O)", params["smiles"], count=1)

    # update the monomer accordingly
    monomer.smiles = params["smiles"]
    monomer.c1_find = lambda x: c1_finder(x, params["smiles"])
    monomer.structure = None
    monomer.get_structure()


def change_base_monomer(monomer, names, types):
    """
    Check if the basic structure of the monomer has been changed

    Args:
        monomer (Monomer): Monomer to be modified according to functional groups
        names (List[str]): List of the names of the functional groups
        types (List[int]): List of type definitions from the recipe of a monosaccharide

    Returns:
        modified monomer object
    """
    longer = get_indices(names, types)[1] is not None
    open_form = len(set(names).intersection({"-ol", "-onic", "-aric", "-ulosonic", "-ulosaric"})) != 0

    if not open_form:
        if not longer:
            # nothing to do
            return monomer

        # parse for sth. like LDManHep or DDManHep -> afterwards, no need to look for LD/DD/... and Hep/Hex/Pen/Oct
        check_for_resizing(monomer, names, types)
        return monomer

    # open form glycan, changed ends and eventually changed length
    check_for_open_form(monomer, names, types, longer)

    return monomer



import copy
import re
from collections import defaultdict

import networkx as nx
import numpy as np
from rdkit import Chem

from glyles.glycans.utils import sanitize_smiles


def masked2nx(smiles: str, mask: list[int]) -> nx.Graph:
    mol = Chem.MolFromSmiles(smiles)
    tmp1, i, read_atom, read_atom_index = defaultdict(list), 0, "", -1
    for i in range(len(smiles)):
        # check for uppercase, i.e., atoms
        if smiles[i].isupper() or "c":
            # save if it's an element, ie starts with uppercase or is a aromatic c
            if read_atom != "" and (read_atom[0].isupper() or read_atom == "c"):
                tmp1[read_atom] += [mask[read_atom_index]]
            # read current atom
            read_atom = smiles[i]
            read_atom_index = i
        # extent current element
        elif smiles[i].isalpha():
            read_atom += smiles[i]
    if read_atom != "" and (read_atom[0].isupper() or read_atom == "c"):
        tmp1[read_atom] += [mask[read_atom_index]]

    atom_counter = defaultdict(int)
    G = nx.Graph()
    for atom in mol.GetAtoms():
        elem = atom.GetSymbol()
        G.add_node(
            atom.GetIdx(),
            atomic_num=elem,
            formal_charge=atom.GetFormalCharge(),
            chiral_tag=atom.GetChiralTag(),
            is_aromatic=atom.GetIsAromatic(),
            mono_id=tmp1[elem][atom_counter[elem]],
        )
        atom_counter[elem] += 1
    for bond in mol.GetBonds():
        start = bond.GetBeginAtomIdx()
        G.add_edge(
            start,
            bond.GetEndAtomIdx(),
            bond_type=bond.GetBondType(),
            mono_id=G.nodes[start]["mono_id"],
        )
    return G


class Merger:
    """
    Merge the tree of monomers into a SMILES representation of the complete molecule.
    """

    def __init__(self, factory):
        """
        Create a merger class to merge a parsed glycan tree into a SMILES string.

        Args:
            factory (MonomerFactory): factory instance to use to generate the monomers for the glycan tree from
        """
        self.factory = factory

    def merge(self, t, root_orientation="n", start=100, smiles_only: bool = True):
        """
        Merge the provided tree of monomers enriched with the glycans in the nodes and information on the bindings
        between two monomer-nodes in the edges. The input graph is not changed during this process, as the methods
        deep-copies the glycan representing datastructure.

        Args:
            t (networkx.DiGraph): Graph representing the glycan to compute the whole SMILES representation for.
            root_orientation (str): Orientation of the root monomer in glycan (can be a (= alpha) or b (= beta))
            start (int): index of the atom to start the SMILES generation from

        Returns:
            SMILES representation as string
        """
        # deep-copy the tree to not modify the provided argument
        t = copy.deepcopy(t)

        # first mark the atoms that will be replaced in a binding of two monomers
        self.mark(t, 0, f"({root_orientation}1-?)")

        # return the string that can be computed from connecting the monomers as marked above
        if np.where(t.nodes[0]["type"].get_features()[:, 1] == start)[0].size != 0:
            position = np.argwhere(t.nodes[0]["type"].get_features()[:, 1] == start).squeeze().item()
        else:
            position = np.argwhere(t.nodes[0]["type"].get_features()[:, 1] == 1).squeeze().item()
        smiles, mask = self.merge_int(t, 0, position, 0)
        if smiles_only:
            return smiles
        return smiles, masked2nx(smiles, mask)

    def mark(self, t, node, p_edge):
        """
        Recursively mark in every node of the molecule which atoms are being replaced by bound monomers.

        Args:
            t (networkx.DiGraph): Tree representing the glycan to compute the whole SMILES representation for.
            node (int): ID of the node to work on in this method
            p_edge (str): edge annotation to parent monomer

        Returns:
            Nothing
        """
        # get children nodes
        children = [x[1] for x in t.edges(node)]

        # set chirality of atom binding parent
        if p_edge is not None and t.nodes[node]["type"].is_non_chiral():
            t.nodes[node]["type"] = t.nodes[node]["type"].to_chirality(p_edge[1], self.factory)

        # check for validity of the tree, ie if it's a leaf (return, nothing to do) or has too many children (Error)
        if len(children) == 0:  # leaf
            return
        if len(children) > 4:  # too many children
            raise NotImplementedError("Glycans with maximal branching factor greater then 3 not implemented.")

        # iterate over the children and the atoms used to mark binding atoms in my structure
        for child, atom in zip(children, t.nodes[node]["type"].get_dummy_atoms()):
            binding = re.findall(r'\d+', t.get_edge_data(node, child)["type"])[1]

            t.nodes[node]["type"].mark(int(binding), *atom)
            self.mark(t, child, t.get_edge_data(node, child)["type"])

    def merge_int(self, t, node, start, ring_index):
        """
        Recursively merge every node of the molecule with its children and get the SMILES representation of the
        subtree.

        Args:
            t (networkx.DiGraph): Graph representing the glycan to compute the whole SMILES representation for.
            node (int): ID of the node to work on in this method
            start (int): ID of the atom in the inner graph to start from when generating the SMILES string
            ring_index (int): Index of the ring to use when generating the SMILES strings

        Returns:
            SMILES representation of the subtree of the given node
        """
        # get my children and compute my SMILES string
        children = [x[1] for x in t.edges(node)]
        me = t.nodes[node]["type"].to_smiles(ring_index, root_id=start)
        mask = [node for _ in range(len(me))]

        # check for validity of the tree, ie if it's a leaf
        if len(children) == 0:  # leaf
            return me, mask

        # iterate over the children and the atoms used to mark binding atoms
        for child, (o_atom, n_atom) in zip(children, t.nodes[node]["type"].get_dummy_atoms()):
            binding = re.findall(r'\d+', t.get_edge_data(node, child)["type"])[0]

            child_start = t.nodes[child]["type"].root_atom_id(int(binding))
            if child_start == -1:
                raise ValueError("No child start found.")

            # get the SMILES of this child and plug it in the current own SMILES
            child_smiles, child_mask = self.merge_int(t, child, child_start, ring_index + 1)
            if o_atom[1] in me:
                m = list(re.finditer(o_atom[2], me))[0]
                me = re.sub(o_atom[2], child_smiles, me)
                mask = mask[:m.start()] + child_mask + mask[m.end():]
            elif n_atom[1] in me:
                m = list(re.finditer(n_atom[2], me))[0]
                me = re.sub(n_atom[2], "N(" + child_smiles[1:] + ")", me)
                mask = mask[:m.start()] + [node, node] + child_mask[1:] + [node] + mask[m.end():]
            me, mask = sanitize_smiles(me, mask)
        return me, mask



# Generated from Glycan.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .GlycanParser import GlycanParser
else:
    from GlycanParser import GlycanParser

# This class defines a complete listener for a parse tree produced by GlycanParser.
class GlycanListener(ParseTreeListener):

    # Enter a parse tree produced by GlycanParser#start.
    def enterStart(self, ctx:GlycanParser.StartContext):
        pass

    # Exit a parse tree produced by GlycanParser#start.
    def exitStart(self, ctx:GlycanParser.StartContext):
        pass


    # Enter a parse tree produced by GlycanParser#begin.
    def enterBegin(self, ctx:GlycanParser.BeginContext):
        pass

    # Exit a parse tree produced by GlycanParser#begin.
    def exitBegin(self, ctx:GlycanParser.BeginContext):
        pass


    # Enter a parse tree produced by GlycanParser#branch.
    def enterBranch(self, ctx:GlycanParser.BranchContext):
        pass

    # Exit a parse tree produced by GlycanParser#branch.
    def exitBranch(self, ctx:GlycanParser.BranchContext):
        pass


    # Enter a parse tree produced by GlycanParser#deriv.
    def enterDeriv(self, ctx:GlycanParser.DerivContext):
        pass

    # Exit a parse tree produced by GlycanParser#deriv.
    def exitDeriv(self, ctx:GlycanParser.DerivContext):
        pass


    # Enter a parse tree produced by GlycanParser#saci.
    def enterSaci(self, ctx:GlycanParser.SaciContext):
        pass

    # Exit a parse tree produced by GlycanParser#saci.
    def exitSaci(self, ctx:GlycanParser.SaciContext):
        pass


    # Enter a parse tree produced by GlycanParser#con.
    def enterCon(self, ctx:GlycanParser.ConContext):
        pass

    # Exit a parse tree produced by GlycanParser#con.
    def exitCon(self, ctx:GlycanParser.ConContext):
        pass


    # Enter a parse tree produced by GlycanParser#add.
    def enterAdd(self, ctx:GlycanParser.AddContext):
        pass

    # Exit a parse tree produced by GlycanParser#add.
    def exitAdd(self, ctx:GlycanParser.AddContext):
        pass


    # Enter a parse tree produced by GlycanParser#fgi.
    def enterFgi(self, ctx:GlycanParser.FgiContext):
        pass

    # Exit a parse tree produced by GlycanParser#fgi.
    def exitFgi(self, ctx:GlycanParser.FgiContext):
        pass


    # Enter a parse tree produced by GlycanParser#carb.
    def enterCarb(self, ctx:GlycanParser.CarbContext):
        pass

    # Exit a parse tree produced by GlycanParser#carb.
    def exitCarb(self, ctx:GlycanParser.CarbContext):
        pass


    # Enter a parse tree produced by GlycanParser#modi.
    def enterModi(self, ctx:GlycanParser.ModiContext):
        pass

    # Exit a parse tree produced by GlycanParser#modi.
    def exitModi(self, ctx:GlycanParser.ModiContext):
        pass


    # Enter a parse tree produced by GlycanParser#qnum.
    def enterQnum(self, ctx:GlycanParser.QnumContext):
        pass

    # Exit a parse tree produced by GlycanParser#qnum.
    def exitQnum(self, ctx:GlycanParser.QnumContext):
        pass


    # Enter a parse tree produced by GlycanParser#typi.
    def enterTypi(self, ctx:GlycanParser.TypiContext):
        pass

    # Exit a parse tree produced by GlycanParser#typi.
    def exitTypi(self, ctx:GlycanParser.TypiContext):
        pass


    # Enter a parse tree produced by GlycanParser#bridge.
    def enterBridge(self, ctx:GlycanParser.BridgeContext):
        pass

    # Exit a parse tree produced by GlycanParser#bridge.
    def exitBridge(self, ctx:GlycanParser.BridgeContext):
        pass


    # Enter a parse tree produced by GlycanParser#ct.
    def enterCt(self, ctx:GlycanParser.CtContext):
        pass

    # Exit a parse tree produced by GlycanParser#ct.
    def exitCt(self, ctx:GlycanParser.CtContext):
        pass



del GlycanParser


from antlr4.error.ErrorListener import ErrorListener
from glyles.glycans.utils import ParseError
import sys


class GlyLESErrorListener(ErrorListener):

    def __init__(self):
        super(GlyLESErrorListener, self).__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        error_msg = f"line {line}:{column} {msg}"
        # Printing to stderr kept to mimic the old behaviour.
        # Maybe not a good idea for a library to pollute the stderr when the exceptions are handled?
        print(error_msg, file=sys.stderr)
        raise ParseError(f"Glycan cannot be parsed:\n{error_msg}")



import os
import sys
import argparse

from glyles import convert
from glyles.version import __version__


def parse_args(args):
    """
    Parse the arguments given to this script.

    Args:
        args: Arguments taken from sys.argv, provided explicitly to be able to test the functionality of this method

    Returns:
        The arguments parsed from the commandline into a dictionary as in kwargs
    """
    parser = argparse.ArgumentParser(
        prog="glyles",
        description="A tool to convert IUPAC representation of Glycans into SMILES representation"
    )
    parser.add_argument(
        '-i',
        '--input',
        required=True,
        action="store",
        help="Input IUPAC string, list or file path",
        nargs='+'
    )
    parser.add_argument(
        '-o',
        '--output',
        required=True,
        action="store",
        help="Output file path"
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}"
    )
    return vars(parser.parse_args(args))


def parse_list(input_list):
    """
    Parse a list of different inputs into a list.

    Args:
        input_list (List[str]): List of IUPAC-condensed strings and filenames to files containing the glycan names

    Returns:
        A list with glycan names taken from the files and the individual glycan names
    """
    glycans = []
    for x in input_list:
        if os.path.isfile(x):
            with open(x, "r") as data:
                glycans += [x.strip() for x in data.readlines()]
        else:
            glycans.append(x)
    return glycans


def main(args):
    """
    Main functionality for the CLI of GlyLES.

    Args:
        args: Arguments taken from sys.argv, provided explicitly to be able to test the functionality of this method
    """
    # parse the arguments
    kwargs = parse_args(args)

    # check if the output already exists
    if os.path.isfile(kwargs["output"]):
        if input(
                "The given output file already exist, do you want to proceed and overwrite it? [y]/n\n"
        ) not in ["", "Y", "y"]:
            exit(0)

    # check if it's a one-element list, in case replace the input with the only element in the input list
    if isinstance(kwargs["input"], list) and len(kwargs["input"]) == 1:
        kwargs["input"] = kwargs["input"][0]

    # forward the input to glyles ...
    if isinstance(kwargs["input"], list):
        data = parse_list(kwargs["input"])
        convert(glycan_list=data, output_file=kwargs["output"])
    elif os.path.isfile(kwargs["input"]):
        convert(glycan_file=kwargs["input"], output_file=kwargs["output"])
    elif isinstance(kwargs["input"], str):
        convert(glycan=kwargs["input"], output_file=kwargs["output"])
    else:
        # ... or raise an error if no matching input type has been found
        print(
            f"An error occurred with the input type: {type(kwargs['input'])}.\n"
            f"Please open an issue on github for this.\n"
            f"https://github.com/kalininalab/GlyLES/issues",
            file=sys.stderr,
        )


if __name__ == "__main__":
    main(sys.argv[1:])



import logging
from collections import Counter
from typing import Union, List

import networkx as nx
from PIL import ImageDraw, Image
from glyles.grammar.GlycanParser import GlycanParser
from networkx.algorithms.isomorphism import DiGraphMatcher
import pydot
from antlr4 import InputStream, CommonTokenStream
from rdkit import Chem
from rdkit.Chem.Descriptors import ExactMolWt

from glyles.glycans.factory.factory import MonomerFactory
from glyles.glycans.mono.reactor import functional_groups
from glyles.glycans.mono.monomer import Monomer
from glyles.glycans.poly.anltr_error_listener import GlyLESErrorListener
from glyles.glycans.poly.merger import Merger
from glyles.glycans.poly.viz import Tree
from glyles.glycans.poly.walker import TreeWalker
from glyles.glycans.utils import ParseError
from glyles.grammar.GlycanLexer import GlycanLexer

if not hasattr(GlycanLexer, "MOD"):
    GlycanLexer.MOD = GlycanLexer.QMARK + 1


def compare_smiles(
        c: Chem.Mol,
        s: Chem.Mol
):
    """
    Compare two molecules if they are equal.

    Args:
        c: molecule 1
        s: molecule 2

    Returns:
        True if their kekulized, canonical SMILES string is equal; False otherwise
    """
    Chem.Kekulize(c)
    Chem.Kekulize(s)

    ssmiles = Chem.MolToSmiles(s, kekuleSmiles=True)
    csmiles = Chem.MolToSmiles(c, kekuleSmiles=True)
    return csmiles == ssmiles


def recipe_equality(
        glycan: Monomer,
        query: Monomer,
        no: bool = False,
        some: bool = False,
        every: bool = False
):
    """
    Checking if two monomers are considered isomorphic under the given mode of isomorphism. One of the bool-flags has
    to be set.

    Note:
        node-match's first argument is always from first node, second argument always from second graph

    Args:
        glycan: monomer to search in
        query: monomer to find in glycan
        no: If True, only match basic monosaccharide, no functional groups
        some: If True, match all of query's groups but not all of glycan's groups
        every: If True, glycans must match exactly, i.e., modifications have to be the same

    Returns:
        True if glycan and query are considered equal under given circumstances
    """
    # check if exactly one isomorphism mode is activated
    if sum([no, some, every]) != 1:
        raise ValueError("Exactly one of arguments no, some, every has to be set to True.")

    if no:
        # if functional groups should not be considered for isomorphism, just check the root monomer of both nodes
        recipe_glycan, recipe_query = glycan.get_recipe(), query.get_recipe()
        return recipe_glycan[list(zip(*recipe_glycan))[1].index(GlycanLexer.SAC)] == \
               recipe_query[list(zip(*recipe_query))[1].index(GlycanLexer.SAC)]
    if some:
        # raise NotImplementedError("Matching to a subset of functional groups is not yet implemented. Coming soon.")
        """# if query's fgs have to be a subset of glycan's fgs, do some magic
        recipe_glycan, recipe_query = glycan.get_recipe(), query.get_recipe()
        if recipe_glycan[list(zip(*recipe_glycan))[1].index(GlycanLexer.SAC)] != \
                recipe_query[list(zip(*recipe_query))[1].index(GlycanLexer.SAC)]:
            return False
        iso = find_isomorphism_nx(glycan, query, query.name, query.c1_find)
        inv = dict([(v, k) for k, v in iso])
        ring_o = query.x[querx.x[:, 1] == 100]
        if ring_o not in inv:
            return False"""
        return len(set(query.get_recipe()).difference(set(glycan.get_recipe()))) == 0
    if every:
        # if all functional groups have to be matched, compare the smiles strings of both
        return compare_smiles(glycan.get_structure(), query.get_structure())
    return False


class Glycan:
    """
    This class is like an interaction with the Parser for the IUPAC representation of the glycan. The grammar for
    glycans is defined using ANTLR (https://www.antlr.org/). From this ANTLR is able to generate lexer and parser that
    fit the defined grammar. Don't touch those files those are auto generated and therefore mostly uncommented.

    The defined grammar discards the last glycan which is used to define the root of the glycan tree. Therefore, the
    resulting abstract syntax trees (AST)s are not intuitive.
    """

    def __init__(self, iupac, root_orientation="n", start=100, tree_only=False, full=True):
        """
        Initialize the glycan from the IUPAC string.

        Args:
            iupac (str): IUPAC string representation of the glycan to represent
            root_orientation (str): orientation of the root monomer in the glycan (choose from 'a', 'b', 'n')
            start (int): ID of the atom to start with in the root monomer when generating the SMILES
            tree_only (int): Flag indicating to only parse the tree of glycans and not the modifications
            full (bool): Flag indicating that only fully convertible glycans should be returned, i.e. all modifications
                such as 3-Anhydro-[...] are also present in the SMILES
        """
        self.iupac = iupac
        self.parse_tree = None
        self.grammar_tree = None
        self.glycan_smiles = None
        self.root_orientation = root_orientation
        self.start = start
        self.tree_only = tree_only
        self.factory = MonomerFactory()
        self.full = full
        self.tree_full = True
        self.__parse()

    def summary(self):
        """
        Aggregate some statistics of the glycan. This includes in the following order [the key in the output dictionary
        in brackets]:
        molecular formula [formula], number of atoms [atoms], number of bonds [bonds], number of rings [rings], number
        of monomers [monomers], max depth of the tree [depth], the root monomer [root], list of all leaf monomers
        [leaves], molecular weight [weight].

        Returns:
            The above named statistics are returned as dictionary with the given keys.
        """
        # generate smiles for this molecule and check it's not empty
        smiles = self.get_smiles()
        if smiles == "":
            raise ValueError("SMILES string for this glycan is empty, check if the IUPAC is convertable.")

        # generate molecule with RDKit and check it's not None
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            raise ValueError("Generated SMILES is invalid, rdkit couldn't read it in.")

        # compute the statistics and return the results inplace
        return {
            "formula": Chem.rdMolDescriptors.CalcMolFormula(mol),
            "weight": ExactMolWt(mol),
            "atoms": len(mol.GetAtoms()),
            "bonds": len(mol.GetBonds()),
            "rings": len(mol.GetRingInfo().AtomRings()),
            "monomers": len(self.parse_tree.nodes),
            "types": dict(Counter([self.parse_tree.nodes[n]["type"].get_name(True) for n in self.parse_tree.nodes])),
            "root": self.parse_tree.nodes[0]["type"].get_name(True),
            "leaves": [self.parse_tree.nodes[n]["type"].get_name(True) for n, d in self.parse_tree.out_degree() if
                       d == 0],
            "depth": max([v for k, v in nx.shortest_path_length(self.parse_tree, 0).items()]),
        }

    def count(self, glycan, match_all_fg=False, match_some_fg=False, match_edges=False, match_nodes=False,
              match_leaves=False, match_root=False):
        """Match a glycan against a query molecule and return the number of hits. This matching can be restricted by
        setting some flags introducing additional conditions of the matches.

        This matching does not include the configuration (alpha/beta/undefined) of the root monomer of the query. So
        query "Gal" will result a hit in "GalNAc6S b" but neither do "Gal a" or "Gal b".

        Args:
            glycan (Union[str, 'Glycan']): query glycan to be matched against the monomers of this glycan
            match_all_fg (bool): flag indicating to match all fgs of the query glycan to all fgs of a monomer
            match_some_fg (bool): flag indicating to match all fgs of the query glycan to some fgs of a monomer
            match_edges (bool): flag indicating to also match edges
            match_nodes (bool): flag indicating to match against all nodes
            match_leaves (bool): flag indicating to match against the leaf monomers only
            match_root (bool): flag indicating to match against the root monomer only

        Returns:
            The number of matches of the query in this glycan under the given conditions
        """
        if sum([match_nodes, match_leaves, match_root]) != 1:
            raise ValueError("Exactly one of match_nodes, match_leaves, match_root has to be True.")

        if isinstance(glycan, str):
            glycan = Glycan(glycan)

        if len(glycan.parse_tree.nodes) != 1 and (match_leaves or match_root):
            raise ValueError("Cannot match polymeric glycan against leaves of glycan. Leaves are monomers.")

        # node-match's first argument is always from first node, second argument always from second graph
        kwargs = {
            "node_match": lambda x, y: recipe_equality(x["type"], y["type"], no=True),
        }
        if match_some_fg:
            kwargs["node_match"] = lambda x, y: recipe_equality(x["type"], y["type"], some=True)
        elif match_all_fg:
            kwargs["node_match"] = lambda x, y: recipe_equality(x["type"], y["type"], every=True)
        if match_edges:
            kwargs["edge_match"] = lambda e, f: e["type"] == f["type"]

        if match_nodes:
            matcher = DiGraphMatcher(self.parse_tree, glycan.parse_tree, **kwargs)
            return len(list(matcher.subgraph_isomorphisms_iter()))
        if match_leaves:
            q = glycan.parse_tree.nodes[0]
            return sum(
                [kwargs["node_match"](self.parse_tree.nodes[n], q) for n, d in self.parse_tree.out_degree() if d == 0])
        if match_root:
            return sum([kwargs["node_match"](self.parse_tree.nodes[0], glycan.parse_tree.nodes[0])])

    def count_protonation(self, grouping):
        """
        Count the possible deprotonation sites in the final molecule.

        Args:
            grouping (bool): If True, count functional groups based on their common atom, so an SO2 group will count as
                1. Otherwise, count groups based on the protonizable oxygen atoms, so an SO2 group will count as 2.

        Returns:
            The number of possible deprotonations in the molecule.
        """
        # generate smiles for this molecule and check it's not empty
        smiles = self.get_smiles()
        if smiles == "":
            raise ValueError("SMILES string for this glycan is empty, check if the IUPAC is convertable.")

        # generate molecule with RDKit and check it's not None
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            raise ValueError("Generated SMILES is invalid, rdkit couldn't read it in.")

        # iterate over deprotonatable functional groups in every form
        count = 0
        for core, group in [
            (6, [("C(O)=O", 1)]),
            (15, [("P(=O)(O)O", 3), ("P(=O)O", 2), ("P=O", 1)]),
            (16, [("S(=O)(=O)O", 2), ("S(=O)(=O)", 1)])
        ]:
            matched_atoms = set()
            for g, val in group:
                # compute the matches against this glycan
                matches = mol.GetSubstructMatches(Chem.MolFromSmiles(g))
                for match in matches:
                    for aid in match:
                        # find the core atom of each match, add it to the list of  covered atoms and increase the count
                        if mol.GetAtomWithIdx(aid).GetAtomicNum() == core and aid not in matched_atoms:
                            matched_atoms.add(aid)
                            # do some trick to either count groups to be deprotonated or possibly chargeable atoms
                            count += val ** (1 - grouping)
        return count

    def count_functional_groups(self, groups):
        """
        Count the number of the provided functional group in the final molecule.

        Args:
            groups (Union[str, List[str]]): each string has to be a valid functional group representable by this tool,
                a valid SMILES string or a valid SMARTS string

        Returns:
            The number of matches of all functional groups. This count might overlap in the matched atoms.
        """
        # in case of string input, put it into a one-element list
        if not isinstance(groups, list):
            groups = [groups]

        # generate smiles for this molecule and check it's not empty
        smiles = self.get_smiles()
        if smiles == "":
            raise ValueError("SMILES string for this glycan is empty, check if the IUPAC is convertable.")

        # generate molecule with RDKit and check it's not None
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            raise ValueError("Generated SMILES is invalid, rdkit couldn't read it in.")

        fgs = []
        for group in groups:
            # convert the functional group into a RDKit molecule
            if group in functional_groups:
                tmp = Chem.MolFromSmiles(functional_groups[group])
            else:
                tmp = Chem.MolFromSmiles(group, sanitize=False)

            # check the functional groups for validity
            if tmp is None:
                raise ValueError(f"The functional group {group} cannot be parsed into a molecule. "
                                 f"Please make sure, it's a valid SMILES string or an implemented functional group.")
            fgs.append(tmp)

        # sum over matches of functional groups
        count = 0
        hydro_mol = Chem.rdmolops.AddHs(mol)
        for g in fgs:
            matched = set()
            if 1 in set(a.GetAtomicNum() for a in g.GetAtoms()):
                matches = hydro_mol.GetSubstructMatches(g)
            else:
                matches = mol.GetSubstructMatches(g)
            for match in matches:
                match = set(match)
                if len(matched.intersection(match)) == 0:
                    count += 1
                    matched = matched.union(match)
        return count

    def get_smiles(self):
        """
        Request the SMILES string of the parsed molecule.

        Returns:
            Generated SMILES string
        """
        # return an empty SMILES if the output is required to represent all modifications, but it actually wouldn't
        if not self.tree_only and self.tree_full != self.full:
            return ""

        if self.glycan_smiles is None:
            self.parse_tree, self.tree_full = TreeWalker(self.factory, False).parse(self.grammar_tree)
            self.glycan_smiles = Merger(self.factory).merge(self.parse_tree, self.root_orientation, start=self.start)
        # self.glycan_smiles = self.glycan_smiles.replace("At", "O-")
        return self.glycan_smiles

    def get_tree(self):
        """
        Request the tree parsed from the IUPAC in this instance.

        Returns:
            The parsed tree with the single monomers in the nodes.
        """
        return self.parse_tree

    def save_dot(self, output, horizontal=False):
        """
        Save the tree structure of the encoded glycan molecule into a dot file visualizing the graph of monomers.

        Args:
            output (str): path to store the DOT file in
            horizontal (bool): Show graph in horizontal orientation from left to right

        Returns:
            pydot graph object containing the graph
        """
        if horizontal:
            graph = pydot.Dot("iupac_tree", rankdir="LR")
        else:
            graph = pydot.Dot("iupac_tree")
        for node in range(len(self.parse_tree.nodes)):
            graph.add_node(pydot.Node(node, label=self.parse_tree.nodes[node]["type"].get_name(full=True)))
        for edge in self.parse_tree.edges():
            graph.add_edge(pydot.Edge(*edge[::-1], label=self.parse_tree.get_edge_data(*edge)["type"]))
        graph.write(output)
        return graph

    def __parse(self):
        """
        Adapter on the Lexer and Parser generated by ANTLR based on glyles/grammar/Glycan.g4.
        """
        try:
            # parse the remaining structure description following the grammar, also add the dummy characters
            if not isinstance(self.iupac, str):
                raise ParseError("Only string input can be parsed: " + str(self.iupac))

            stream = InputStream(data='#' + self.iupac + '#')
            lexer = GlycanLexer(stream)
            token = CommonTokenStream(lexer)
            parser = GlycanParser(token)

            lexer.removeErrorListeners()
            lexer.addErrorListener(GlyLESErrorListener())

            parser.removeErrorListeners()
            parser.addErrorListener(GlyLESErrorListener())

            try:
                self.grammar_tree = parser.start()
            except ParseError as pe:
                self.parse_tree = None
                self.glycan_smiles = ""
                raise pe from None

            # walk through the AST and parse the AST into a networkx representation of the glycan.
            self.parse_tree, self.tree_full = TreeWalker(self.factory, self.tree_only).parse(self.grammar_tree)

            # if the glycan should be parsed immediately, do so
            if not self.tree_only and self.tree_full and self.full:
                self.glycan_smiles = Merger(self.factory).merge(self.parse_tree, self.root_orientation, start=self.start)
                # catch any exception at glycan level to not destroy the whole pipeline because of one mis-formed glycan
                print(self.glycan_smiles)

        except ParseError as e:
            msg = e.__str__().replace("\n", " ")
            logging.error(f"A parsing error occurred with \"{self.iupac}\": Error message: {msg}")
            self.glycan_smiles = ""
        except Exception as e:
            msg = e.__str__().replace("\n", " ")
            logging.error(f"An unexpected exception occurred with \"{self.iupac}\". This glycan cannot be parsed. "
                          f"Error message: {msg}")
            self.glycan_smiles = ""
            raise e



import networkx as nx
from antlr4 import ErrorNode, TerminalNode

from glyles.glycans.utils import UnreachableError, ketoses2
from glyles.grammar.GlycanLexer import GlycanLexer
from glyles.grammar.GlycanParser import GlycanParser

if not hasattr(GlycanLexer, "MOD"):
    GlycanLexer.MOD = GlycanLexer.QMARK + 1


class TreeWalker:
    def __init__(self, factory, tree_only):
        """
        Just initialize an empty tree.

        Args:
            factory (MonomerFactory): factory instance to use to generate the monomers for the glycan tree from
        tree_only (bool): Flag indicating to only parse the tree of glycans and not the modifications
        """
        self.g = nx.DiGraph()
        self.factory = factory
        self.node_id = 0
        self.tree_only = tree_only
        self.full = True

    def parse(self, t):
        for c in filter(lambda x: not isinstance(x, (TerminalNode, ErrorNode)), t.getChildren()):
            if isinstance(c, GlycanParser.BranchContext):
                self.__walk(c, self.node_id)
            elif isinstance(c, GlycanParser.BeginContext):
                self.__parse(c)
        return self.g, self.full and len(list(nx.connected_components(self.g.to_undirected()))) == 1

    def __parse(self, t):
        """
        Parse a parsed tree (AST) from ANTLR into this networkx graph.

        Args:
            t (GlycanParser.StartContext): result of the parsing step from ANTLR

        Returns:
            Tree of parsed glycan with monomers in nodes
        """

        # parse the initial monomer and the orientation of the root monomer
        # and remove the first and last char, i.e. '{' and '}'
        # if isinstance(t, )
        children = list(t.getChildren())
        if len(children) == 1:  # glycan
            self.__add_node(children[0])
        elif len(children) == 2:  # branch glycan
            node_id = self.__add_node(children[1])
            self.__walk(children[0], node_id)
        elif len(children) == 3:  # SAC ' ' TYPE
            self.__add_node(children[0], children[2].symbol.text)
        elif len(children) == 4:  # branch SAC ' ' TYPE
            node_id = self.__add_node(children[1], children[3].symbol.text)
            self.__walk(children[0], node_id)
        else:
            raise RuntimeError("This branch of the if-statement should be unreachable!")

    def __walk(self, t, parent):
        """
        The heart of this class. This method recursively adds nodes to the graph and connects them according to the
        connections in the input IUPAC.

        Args:
            t (GlycanParser.StartContext): subtree to parse in this recursive step.
            parent (int): ID of the parent node for this (subtree)

        Returns:
            ID of the parent for the remaining recursive procedure after resolving this subtree
        """
        # security check for some nodes that should not occur, but who knows what ANTLR does (or does not) ;-)
        if isinstance(t, ErrorNode):
            raise UnreachableError("ErrorNodes in parsing!")
        elif isinstance(t, TerminalNode):
            raise UnreachableError("TerminalNodes in parsing!")

        children = list(t.getChildren())
        if len(children) == 2:  # {glycan con}
            # terminal element, add the node with the connection
            node_id = self.__add_node(children[0])
            self.full &= self.__add_edge(parent, node_id, children[1])
            return node_id

        elif len(children) == 3 and isinstance(children[2], GlycanParser.BranchContext):  # {glycan con branch}
            # chain without branching, the parent is the parent of the parsing of the back part
            parent = self.__walk(children[2], parent)
            node_id = self.__add_node(children[0])
            self.full &= self.__add_edge(parent, node_id, children[1])
            return node_id

        elif len(children) == 3 and isinstance(children[1], GlycanParser.BranchContext):  # {'[' branch ']'}
            # branching, hand the parent on to the next level
            self.__walk(children[1], parent)
            return parent

        elif len(children) == 6:  # {glycan con '[' branch ']' branch}
            # branching in a chain, append the end to the parent and hang both branches on that
            node_id = self.__walk(children[5], parent)
            self.__walk(children[3], node_id)
            node_id2 = self.__add_node(children[0])
            self.full &= self.__add_edge(node_id, node_id2, children[1])
            return node_id2

        elif len(children) == 9:  # {glycan con '[' branch ']' '[' branch ']' branch}
            # branching in a chain, append the end to the parent and hang both branches on that
            node_id = self.__walk(children[8], parent)
            self.__walk(children[3], node_id)
            self.__walk(children[6], node_id)
            node_id2 = self.__add_node(children[0])
            self.full &= self.__add_edge(node_id, node_id2, children[1])
            return node_id2

        elif len(children) == 12:  # {glycan con '[' branch ']' '[' branch ']' '[' branch ']' branch}
            # branching in a chain, append the end to the parent and hang both branches on that
            node_id = self.__walk(children[11], parent)
            self.__walk(children[3], node_id)
            self.__walk(children[6], node_id)
            self.__walk(children[9], node_id)
            node_id2 = self.__add_node(children[0])
            self.full &= self.__add_edge(node_id, node_id2, children[1])
            return node_id2

        # there should be no case missing, but who knows...
        raise UnreachableError("Invalid branching in glycan tree")

    def context2str(self, node):
        output = ""
        for c in node.getChildren():
            if isinstance(c, TerminalNode):
                output += str(c)
            else:
                output += self.context2str(c)
        return output

    def __add_edge(self, parent, child, con) -> bool:
        """
        Add an edge between the provided ids of the parent and the children in the glycan tree.

        Args:
            parent (int): ID of parent in the connection of the glycan tree
            child (int): ID of the child in the connection of the glycan tree
            con (str): Connection element from the ParseTree

        Returns:
            Nothing
        """
        # for floating elements as they add a new connected component
        if parent == child:
            return True

        con = self.context2str(con)
        if "(" not in con and ")" not in con:
            if "-" not in con:
                bond = ("2-" if (self.g.nodes[child]["type"].get_lactole,
                                 self.g.nodes[child]['type'].get_name()) in ketoses2 else "1-")
                con = con[0] + bond + con[1:]
            con = "(" + con + ")"
        self.g.add_edge(parent, child, type=con)
        # Identify standard and narrow wildcard connections
        return "?" not in con and "/" not in con

    def build_recipe(self, node):
        recipe = []
        for c in node.getChildren():
            if isinstance(c, TerminalNode):
                recipe.append((str(c), GlycanLexer.SAC if isinstance(node, GlycanParser.SaciContext) else c.symbol.type))
            else:
                tmp = self.build_recipe(c)
                if isinstance(c, GlycanParser.ModiContext):
                    recipe.append(("".join([x[0] for x in tmp]), GlycanLexer.MOD))
                elif isinstance(c, GlycanParser.SaciContext):
                    recipe += [(x[0], GlycanLexer.SAC) for x in tmp]
                else:
                    recipe += tmp
        return recipe

    def __add_node(self, node, config=""):
        """
        Add a new node to the network based on the name of the represented glycan.

        Args:
            node (GlycanParser.GlycanContext): Node of the parsed tree to be parsed into a monomer
            config (str): configuration to be applied to the monomer

        Returns:
            ID of the newly added node
        """
        # get the id for the node and increase the id for the next node
        node_id = self.node_id
        self.node_id += 1

        # add the node to the network and store the enum of the glycan as attribute
        recipe = self.build_recipe(node)
        # print(recipe)
        monomer, full = self.factory.create(recipe, config, tree_only=self.tree_only)
        self.g.add_node(
            node_id,
            type=monomer,
        )
        self.full &= full

        return node_id



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glyles',
 'glyles.glycans',
 'glyles.glycans.factory',
 'glyles.glycans.mono',
 'glyles.glycans.poly',
 'glyles.grammar']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime==4.13.1',
 'joblib>=1.2.0,<2.0.0',
 'networkx>=2.6.3',
 'numpy',
 'pydot>=1.4.2,<2.0.0',
 'rdkit>=2021.9.2']

entry_points = \
{'console_scripts': ['glyles = glyles.__main__:main']}

setup_kwargs = {
    'name': 'glyles',
    'version': '1.1.0',
    'description': 'A tool to convert IUPAC representation of glycans into SMILES strings',
    'long_description': '# GlyLES\n\n![testing](https://github.com/kalininalab/glyles/actions/workflows/test.yaml/badge.svg)\n[![docs-image](https://readthedocs.org/projects/glyles/badge/?version=latest)](https://glyles.readthedocs.io/en/latest/)\n[![piwheels](https://img.shields.io/piwheels/v/glyles)](https://pypi.org/project/glyles/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/glyles)](https://pypi.org/project/glyles/)\n[![codecov](https://codecov.io/gh/kalininalab/GlyLES/branch/main/graph/badge.svg)](https://codecov.io/gh/kalininalab/glyles)\n[![DOI](https://zenodo.org/badge/431874597.svg)](https://zenodo.org/badge/latestdoi/431874597)\n\nA tool to convert IUPAC representation of Glycans into SMILES representation. This repo is still in the development \nphase; so, feel free to report any errors or issues. The code is available on \n[github](https://github.com/kalininalab/GlyLES/) and the documentation can be found on \n[ReadTheDocs](https://glyles.readthedocs.io/en/latest/index.html).\n\n## Specification and (current) Limitations\n\nThe exact specification we\'re referring to when talking about "IUPAC representations of glycan" or "IUPAC-condensed", \nis given in the "Notes" section of this [website](https://www.ncbi.nlm.nih.gov/glycans/snfg.html). But as this package \nis still in the development phase, not everything of the specification is implemented yet (especially not all side \nchains you can attach to monomers). The structure of the glycan can be represented as a tree of the monosaccharides \nwith maximal branching factor 4, i.e., each monomer in the glycan has at most 4 children.\n\n## Installation\n\nSo far, this package can only be downloaded from the python package index. So the installation with `pip` is very easy.\nJust type\n\n``````shell\npip install glyles\n``````\n\nand you\'re ready to use it as described below. Use \n\n``````shell\npip install --upgrade glyles\n``````\n\nto upgrade the glyles package to the most recent version.\n\n## Basic Usage\n\n### As a Python Package\n\nConvert the IUPAC into a SMILES representation using the handy `convert` method\n\n``````python\nfrom glyles import convert\n\nconvert(glycan="Man(a1-2)Man", output_file="./test.txt")\n``````\n\nYou can also use the `convert_generator` method to get a generator for all SMILES:\n\n``````python\nfrom glyles import convert_generator\n\nfor smiles in convert_generator(glycan_list=["Man(a1-2)Man a", "Man(a1-2)Man b"]):\n    print(smiles)\n``````\n\nFor more examples of how to use this package, please see the notebooks in the \n[examples](https://github.com/kalininalab/GlyLES/tree/dev/examples) folder and checkout the documentation on \n[ReadTheDocs](https://glyles.readthedocs.io/en/latest/index.html).\n\n### In the Commandline\n\nAs of version 0.5.9, there is a commandline interface to GlyLES which is automatically installed when installing GlyLES \nthrough pip. The CLI is open for one or multiple IUPAC inputs as individual arguments. Due to the syntax of the \nIUPAC-condensed notation and the argument parsing in commandlines, the IUPAC strings must be given in quotes.\n\n``````shell\nglyles -i "Man(a1-2)Man" -o test_output.txt\nglyles -i "Man(a1-2)Man" "Fuc(a1-6)Glc" -o test_output.txt\n``````\n\nFile-input is also possible.\n``````shell\nglyles -i input_file.txt -o test_output.txt\n``````\n\nProviding multiple files and IUPAC-condensed names is als supported.\n``````shell\nglyles -i input_file1.txt "Man(a1-2)Man" input_file2.txt input_file13.txt "Fuc(a1-6)Glc" -o test_output.txt\n``````\n\n## Notation of glycans\n\nThere are multiple different notations for glycans in IUPAC. So, according to the \n[SNGF specification](https://www.ncbi.nlm.nih.gov/glycans/snfg.html), `Man(a1-4)Gal`, `Mana1-4Gal`, and `Mana4Gal` \nall describe the same disaccharide. This is also covered in this package as all three notations will be parsed into the \nsame tree of monosaccharides and result in the same SMILES string.\n\nThis is also described more detailed in a section on [ReadTheDocs](https://glyles.readthedocs.io/en/latest/notes/notation.html).\n\n## Poetry\n\nTo develop this package, we use the poetry package manager (see [here](https://python-poetry.org/) for detailed\ninstruction). It has basically the same functionality as conda but supports the package management better and also \nsupports distinguishing packages into those that are needed to use the package and those that are needed in the \ndevelopment of the package. To enable others to work on this repository, we also publish the exact \nspecifications of our poetry environment.\n\n## Citation\n\nIf you use GlyLES in your work, please cite\n```\n@article{joeres2023glyles,\n  title={GlyLES: Grammar-based Parsing of Glycans from IUPAC-condensed to SMILES},\n  author={Joeres, Roman and Bojar, Daniel and Kalinina, Olga V},\n  journal={Journal of Cheminformatics},\n  volume={15},\n  number={1},\n  pages={1--11},\n  year={2023},\n  publisher={BioMed Central}\n}\n```\n',
    'author': 'Roman Joeres',
    'author_email': 'roman.joeres@helmholtz-hips.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kalininalab/GlyLES',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Generated from Glycan.g4 by ANTLR 4.13.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,34,262,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        1,0,1,0,1,0,1,0,1,0,5,0,34,8,0,10,0,12,0,37,9,0,1,0,1,0,1,0,1,1,
        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,55,8,1,1,2,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,3,2,98,8,2,1,3,5,3,101,8,3,10,3,12,
        3,104,9,3,1,3,4,3,107,8,3,11,3,12,3,108,1,3,5,3,112,8,3,10,3,12,
        3,115,9,3,1,3,3,3,118,8,3,1,3,5,3,121,8,3,10,3,12,3,124,9,3,1,3,
        3,3,127,8,3,1,4,1,4,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,
        1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,3,5,152,8,5,1,6,1,6,1,6,
        3,6,157,8,6,1,6,1,6,1,6,3,6,162,8,6,1,6,5,6,165,8,6,10,6,12,6,168,
        9,6,1,6,1,6,1,6,1,6,1,6,1,6,5,6,176,8,6,10,6,12,6,179,9,6,1,6,3,
        6,182,8,6,1,7,1,7,1,7,3,7,187,8,7,1,8,3,8,190,8,8,1,8,1,8,1,8,5,
        8,195,8,8,10,8,12,8,198,9,8,1,9,1,9,3,9,202,8,9,1,9,1,9,1,9,1,9,
        1,9,1,9,1,9,1,9,1,9,1,9,1,9,1,9,3,9,216,8,9,1,9,3,9,219,8,9,1,9,
        5,9,222,8,9,10,9,12,9,225,9,9,1,9,1,9,1,9,1,9,3,9,231,8,9,1,9,1,
        9,1,9,3,9,236,8,9,1,9,1,9,1,9,1,9,1,9,3,9,243,8,9,1,10,1,10,1,10,
        1,10,5,10,249,8,10,10,10,12,10,252,9,10,3,10,254,8,10,1,11,1,11,
        1,12,1,12,1,13,1,13,1,13,0,0,14,0,2,4,6,8,10,12,14,16,18,20,22,24,
        26,0,5,2,0,3,3,13,13,2,0,25,26,32,32,2,0,14,14,34,34,1,0,4,7,2,0,
        27,27,30,30,289,0,28,1,0,0,0,2,54,1,0,0,0,4,97,1,0,0,0,6,102,1,0,
        0,0,8,128,1,0,0,0,10,151,1,0,0,0,12,181,1,0,0,0,14,186,1,0,0,0,16,
        189,1,0,0,0,18,242,1,0,0,0,20,253,1,0,0,0,22,255,1,0,0,0,24,257,
        1,0,0,0,26,259,1,0,0,0,28,35,5,1,0,0,29,30,5,23,0,0,30,31,3,4,2,
        0,31,32,5,24,0,0,32,34,1,0,0,0,33,29,1,0,0,0,34,37,1,0,0,0,35,33,
        1,0,0,0,35,36,1,0,0,0,36,38,1,0,0,0,37,35,1,0,0,0,38,39,3,2,1,0,
        39,40,5,1,0,0,40,1,1,0,0,0,41,42,3,4,2,0,42,43,3,6,3,0,43,44,5,2,
        0,0,44,45,5,14,0,0,45,55,1,0,0,0,46,47,3,4,2,0,47,48,3,6,3,0,48,
        55,1,0,0,0,49,50,3,6,3,0,50,51,5,2,0,0,51,52,5,14,0,0,52,55,1,0,
        0,0,53,55,3,6,3,0,54,41,1,0,0,0,54,46,1,0,0,0,54,49,1,0,0,0,54,53,
        1,0,0,0,55,3,1,0,0,0,56,57,3,6,3,0,57,58,3,10,5,0,58,98,1,0,0,0,
        59,60,3,6,3,0,60,61,3,10,5,0,61,62,3,4,2,0,62,98,1,0,0,0,63,64,5,
        21,0,0,64,65,3,4,2,0,65,66,5,22,0,0,66,98,1,0,0,0,67,68,3,6,3,0,
        68,69,3,10,5,0,69,70,5,21,0,0,70,71,3,4,2,0,71,72,5,22,0,0,72,73,
        3,4,2,0,73,98,1,0,0,0,74,75,3,6,3,0,75,76,3,10,5,0,76,77,5,21,0,
        0,77,78,3,4,2,0,78,79,5,22,0,0,79,80,5,21,0,0,80,81,3,4,2,0,81,82,
        5,22,0,0,82,83,3,4,2,0,83,98,1,0,0,0,84,85,3,6,3,0,85,86,3,10,5,
        0,86,87,5,21,0,0,87,88,3,4,2,0,88,89,5,22,0,0,89,90,5,21,0,0,90,
        91,3,4,2,0,91,92,5,22,0,0,92,93,5,21,0,0,93,94,3,4,2,0,94,95,5,22,
        0,0,95,96,3,4,2,0,96,98,1,0,0,0,97,56,1,0,0,0,97,59,1,0,0,0,97,63,
        1,0,0,0,97,67,1,0,0,0,97,74,1,0,0,0,97,84,1,0,0,0,98,5,1,0,0,0,99,
        101,3,18,9,0,100,99,1,0,0,0,101,104,1,0,0,0,102,100,1,0,0,0,102,
        103,1,0,0,0,103,106,1,0,0,0,104,102,1,0,0,0,105,107,3,8,4,0,106,
        105,1,0,0,0,107,108,1,0,0,0,108,106,1,0,0,0,108,109,1,0,0,0,109,
        113,1,0,0,0,110,112,3,18,9,0,111,110,1,0,0,0,112,115,1,0,0,0,113,
        111,1,0,0,0,113,114,1,0,0,0,114,117,1,0,0,0,115,113,1,0,0,0,116,
        118,5,15,0,0,117,116,1,0,0,0,117,118,1,0,0,0,118,122,1,0,0,0,119,
        121,3,18,9,0,120,119,1,0,0,0,121,124,1,0,0,0,122,120,1,0,0,0,122,
        123,1,0,0,0,123,126,1,0,0,0,124,122,1,0,0,0,125,127,5,14,0,0,126,
        125,1,0,0,0,126,127,1,0,0,0,127,7,1,0,0,0,128,129,7,0,0,0,129,9,
        1,0,0,0,130,131,5,19,0,0,131,132,3,22,11,0,132,133,5,16,0,0,133,
        134,5,18,0,0,134,135,3,20,10,0,135,136,5,20,0,0,136,152,1,0,0,0,
        137,138,5,19,0,0,138,139,5,16,0,0,139,140,5,18,0,0,140,141,3,20,
        10,0,141,142,5,20,0,0,142,152,1,0,0,0,143,144,3,22,11,0,144,145,
        5,16,0,0,145,146,5,18,0,0,146,147,3,20,10,0,147,152,1,0,0,0,148,
        149,3,22,11,0,149,150,5,16,0,0,150,152,1,0,0,0,151,130,1,0,0,0,151,
        137,1,0,0,0,151,143,1,0,0,0,151,148,1,0,0,0,152,11,1,0,0,0,153,154,
        5,31,0,0,154,156,5,23,0,0,155,157,3,26,13,0,156,155,1,0,0,0,156,
        157,1,0,0,0,157,158,1,0,0,0,158,166,5,16,0,0,159,161,5,17,0,0,160,
        162,3,26,13,0,161,160,1,0,0,0,161,162,1,0,0,0,162,163,1,0,0,0,163,
        165,5,16,0,0,164,159,1,0,0,0,165,168,1,0,0,0,166,164,1,0,0,0,166,
        167,1,0,0,0,167,169,1,0,0,0,168,166,1,0,0,0,169,182,5,24,0,0,170,
        171,5,27,0,0,171,172,5,23,0,0,172,177,5,16,0,0,173,174,5,17,0,0,
        174,176,5,16,0,0,175,173,1,0,0,0,176,179,1,0,0,0,177,175,1,0,0,0,
        177,178,1,0,0,0,178,180,1,0,0,0,179,177,1,0,0,0,180,182,5,24,0,0,
        181,153,1,0,0,0,181,170,1,0,0,0,182,13,1,0,0,0,183,187,5,13,0,0,
        184,187,3,24,12,0,185,187,5,8,0,0,186,183,1,0,0,0,186,184,1,0,0,
        0,186,185,1,0,0,0,187,15,1,0,0,0,188,190,7,1,0,0,189,188,1,0,0,0,
        189,190,1,0,0,0,190,191,1,0,0,0,191,192,5,4,0,0,192,196,5,16,0,0,
        193,195,3,12,6,0,194,193,1,0,0,0,195,198,1,0,0,0,196,194,1,0,0,0,
        196,197,1,0,0,0,197,17,1,0,0,0,198,196,1,0,0,0,199,200,5,16,0,0,
        200,202,5,17,0,0,201,199,1,0,0,0,201,202,1,0,0,0,202,203,1,0,0,0,
        203,204,5,16,0,0,204,205,5,18,0,0,205,206,5,9,0,0,206,243,5,18,0,
        0,207,208,5,16,0,0,208,209,5,18,0,0,209,210,3,24,12,0,210,211,5,
        18,0,0,211,212,3,14,7,0,212,213,5,18,0,0,213,243,1,0,0,0,214,216,
        5,18,0,0,215,214,1,0,0,0,215,216,1,0,0,0,216,218,1,0,0,0,217,219,
        5,16,0,0,218,217,1,0,0,0,218,219,1,0,0,0,219,223,1,0,0,0,220,222,
        3,24,12,0,221,220,1,0,0,0,222,225,1,0,0,0,223,221,1,0,0,0,223,224,
        1,0,0,0,224,226,1,0,0,0,225,223,1,0,0,0,226,243,3,14,7,0,227,235,
        5,16,0,0,228,229,5,17,0,0,229,231,5,16,0,0,230,228,1,0,0,0,230,231,
        1,0,0,0,231,232,1,0,0,0,232,236,5,28,0,0,233,236,5,29,0,0,234,236,
        3,16,8,0,235,230,1,0,0,0,235,233,1,0,0,0,235,234,1,0,0,0,236,243,
        1,0,0,0,237,243,5,10,0,0,238,239,5,11,0,0,239,243,5,18,0,0,240,241,
        5,18,0,0,241,243,5,12,0,0,242,201,1,0,0,0,242,207,1,0,0,0,242,215,
        1,0,0,0,242,227,1,0,0,0,242,237,1,0,0,0,242,238,1,0,0,0,242,240,
        1,0,0,0,243,19,1,0,0,0,244,254,5,34,0,0,245,250,5,16,0,0,246,247,
        5,33,0,0,247,249,5,16,0,0,248,246,1,0,0,0,249,252,1,0,0,0,250,248,
        1,0,0,0,250,251,1,0,0,0,251,254,1,0,0,0,252,250,1,0,0,0,253,244,
        1,0,0,0,253,245,1,0,0,0,254,21,1,0,0,0,255,256,7,2,0,0,256,23,1,
        0,0,0,257,258,7,3,0,0,258,25,1,0,0,0,259,260,7,4,0,0,260,27,1,0,
        0,0,27,35,54,97,102,108,113,117,122,126,151,156,161,166,177,181,
        186,189,196,201,215,218,223,230,235,242,250,253
    ]

class GlycanParser ( Parser ):

    grammarFileName = "Glycan.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'#'", "' '", "<INVALID>", "'C'", "'N'", 
                     "'O'", "'P'", "<INVALID>", "'Anhydro'", "'0d'", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "','", "'-'", "'('", "')'", "'['", "']'", 
                     "'{'", "'}'", "'a'", "'ai'", "'c'", "'d'", "'e'", "'t'", 
                     "'='", "'i'", "'/'", "'?'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "SAC", "CARBON", 
                      "NITROGEN", "OXYGEN", "PHOSPHOR", "FG", "ANHYDRO", 
                      "HEAD", "HEADD", "END", "COUNT", "TYPE", "RING", "NUM", 
                      "COLON", "DASH", "LPAR", "RPAR", "LBRACE", "RBRACE", 
                      "LBRACK", "RBRACK", "A", "AI", "C", "D", "E", "T", 
                      "EQ", "I", "SLASH", "QMARK" ]

    RULE_start = 0
    RULE_begin = 1
    RULE_branch = 2
    RULE_deriv = 3
    RULE_saci = 4
    RULE_con = 5
    RULE_add = 6
    RULE_fgi = 7
    RULE_carb = 8
    RULE_modi = 9
    RULE_qnum = 10
    RULE_typi = 11
    RULE_bridge = 12
    RULE_ct = 13

    ruleNames =  [ "start", "begin", "branch", "deriv", "saci", "con", "add", 
                   "fgi", "carb", "modi", "qnum", "typi", "bridge", "ct" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    SAC=3
    CARBON=4
    NITROGEN=5
    OXYGEN=6
    PHOSPHOR=7
    FG=8
    ANHYDRO=9
    HEAD=10
    HEADD=11
    END=12
    COUNT=13
    TYPE=14
    RING=15
    NUM=16
    COLON=17
    DASH=18
    LPAR=19
    RPAR=20
    LBRACE=21
    RBRACE=22
    LBRACK=23
    RBRACK=24
    A=25
    AI=26
    C=27
    D=28
    E=29
    T=30
    EQ=31
    I=32
    SLASH=33
    QMARK=34

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class StartContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def begin(self):
            return self.getTypedRuleContext(GlycanParser.BeginContext,0)


        def LBRACK(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.LBRACK)
            else:
                return self.getToken(GlycanParser.LBRACK, i)

        def branch(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.BranchContext)
            else:
                return self.getTypedRuleContext(GlycanParser.BranchContext,i)


        def RBRACK(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.RBRACK)
            else:
                return self.getToken(GlycanParser.RBRACK, i)

        def getRuleIndex(self):
            return GlycanParser.RULE_start

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStart" ):
                listener.enterStart(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStart" ):
                listener.exitStart(self)




    def start(self):

        localctx = GlycanParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            self.match(GlycanParser.T__0)
            self.state = 35
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==23:
                self.state = 29
                self.match(GlycanParser.LBRACK)
                self.state = 30
                self.branch()
                self.state = 31
                self.match(GlycanParser.RBRACK)
                self.state = 37
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 38
            self.begin()
            self.state = 39
            self.match(GlycanParser.T__0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BeginContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def branch(self):
            return self.getTypedRuleContext(GlycanParser.BranchContext,0)


        def deriv(self):
            return self.getTypedRuleContext(GlycanParser.DerivContext,0)


        def TYPE(self):
            return self.getToken(GlycanParser.TYPE, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_begin

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBegin" ):
                listener.enterBegin(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBegin" ):
                listener.exitBegin(self)




    def begin(self):

        localctx = GlycanParser.BeginContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_begin)
        try:
            self.state = 54
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 41
                self.branch()
                self.state = 42
                self.deriv()
                self.state = 43
                self.match(GlycanParser.T__1)
                self.state = 44
                self.match(GlycanParser.TYPE)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 46
                self.branch()
                self.state = 47
                self.deriv()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 49
                self.deriv()
                self.state = 50
                self.match(GlycanParser.T__1)
                self.state = 51
                self.match(GlycanParser.TYPE)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 53
                self.deriv()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BranchContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def deriv(self):
            return self.getTypedRuleContext(GlycanParser.DerivContext,0)


        def con(self):
            return self.getTypedRuleContext(GlycanParser.ConContext,0)


        def branch(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.BranchContext)
            else:
                return self.getTypedRuleContext(GlycanParser.BranchContext,i)


        def LBRACE(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.LBRACE)
            else:
                return self.getToken(GlycanParser.LBRACE, i)

        def RBRACE(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.RBRACE)
            else:
                return self.getToken(GlycanParser.RBRACE, i)

        def getRuleIndex(self):
            return GlycanParser.RULE_branch

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBranch" ):
                listener.enterBranch(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBranch" ):
                listener.exitBranch(self)




    def branch(self):

        localctx = GlycanParser.BranchContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_branch)
        try:
            self.state = 97
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 56
                self.deriv()
                self.state = 57
                self.con()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 59
                self.deriv()
                self.state = 60
                self.con()
                self.state = 61
                self.branch()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 63
                self.match(GlycanParser.LBRACE)
                self.state = 64
                self.branch()
                self.state = 65
                self.match(GlycanParser.RBRACE)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 67
                self.deriv()
                self.state = 68
                self.con()
                self.state = 69
                self.match(GlycanParser.LBRACE)
                self.state = 70
                self.branch()
                self.state = 71
                self.match(GlycanParser.RBRACE)
                self.state = 72
                self.branch()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 74
                self.deriv()
                self.state = 75
                self.con()
                self.state = 76
                self.match(GlycanParser.LBRACE)
                self.state = 77
                self.branch()
                self.state = 78
                self.match(GlycanParser.RBRACE)
                self.state = 79
                self.match(GlycanParser.LBRACE)
                self.state = 80
                self.branch()
                self.state = 81
                self.match(GlycanParser.RBRACE)
                self.state = 82
                self.branch()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 84
                self.deriv()
                self.state = 85
                self.con()
                self.state = 86
                self.match(GlycanParser.LBRACE)
                self.state = 87
                self.branch()
                self.state = 88
                self.match(GlycanParser.RBRACE)
                self.state = 89
                self.match(GlycanParser.LBRACE)
                self.state = 90
                self.branch()
                self.state = 91
                self.match(GlycanParser.RBRACE)
                self.state = 92
                self.match(GlycanParser.LBRACE)
                self.state = 93
                self.branch()
                self.state = 94
                self.match(GlycanParser.RBRACE)
                self.state = 95
                self.branch()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DerivContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def modi(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.ModiContext)
            else:
                return self.getTypedRuleContext(GlycanParser.ModiContext,i)


        def saci(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.SaciContext)
            else:
                return self.getTypedRuleContext(GlycanParser.SaciContext,i)


        def RING(self):
            return self.getToken(GlycanParser.RING, 0)

        def TYPE(self):
            return self.getToken(GlycanParser.TYPE, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_deriv

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDeriv" ):
                listener.enterDeriv(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDeriv" ):
                listener.exitDeriv(self)




    def deriv(self):

        localctx = GlycanParser.DerivContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_deriv)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 102
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,3,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 99
                    self.modi() 
                self.state = 104
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,3,self._ctx)

            self.state = 106 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 105
                    self.saci()

                else:
                    raise NoViableAltException(self)
                self.state = 108 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,4,self._ctx)

            self.state = 113
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,5,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 110
                    self.modi() 
                self.state = 115
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,5,self._ctx)

            self.state = 117
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==15:
                self.state = 116
                self.match(GlycanParser.RING)


            self.state = 122
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 339440) != 0):
                self.state = 119
                self.modi()
                self.state = 124
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 126
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
            if la_ == 1:
                self.state = 125
                self.match(GlycanParser.TYPE)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SaciContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COUNT(self):
            return self.getToken(GlycanParser.COUNT, 0)

        def SAC(self):
            return self.getToken(GlycanParser.SAC, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_saci

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSaci" ):
                listener.enterSaci(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSaci" ):
                listener.exitSaci(self)




    def saci(self):

        localctx = GlycanParser.SaciContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_saci)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 128
            _la = self._input.LA(1)
            if not(_la==3 or _la==13):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ConContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LPAR(self):
            return self.getToken(GlycanParser.LPAR, 0)

        def typi(self):
            return self.getTypedRuleContext(GlycanParser.TypiContext,0)


        def NUM(self):
            return self.getToken(GlycanParser.NUM, 0)

        def DASH(self):
            return self.getToken(GlycanParser.DASH, 0)

        def qnum(self):
            return self.getTypedRuleContext(GlycanParser.QnumContext,0)


        def RPAR(self):
            return self.getToken(GlycanParser.RPAR, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_con

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCon" ):
                listener.enterCon(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCon" ):
                listener.exitCon(self)




    def con(self):

        localctx = GlycanParser.ConContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_con)
        try:
            self.state = 151
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,9,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 130
                self.match(GlycanParser.LPAR)
                self.state = 131
                self.typi()
                self.state = 132
                self.match(GlycanParser.NUM)
                self.state = 133
                self.match(GlycanParser.DASH)
                self.state = 134
                self.qnum()
                self.state = 135
                self.match(GlycanParser.RPAR)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 137
                self.match(GlycanParser.LPAR)
                self.state = 138
                self.match(GlycanParser.NUM)
                self.state = 139
                self.match(GlycanParser.DASH)
                self.state = 140
                self.qnum()
                self.state = 141
                self.match(GlycanParser.RPAR)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 143
                self.typi()
                self.state = 144
                self.match(GlycanParser.NUM)
                self.state = 145
                self.match(GlycanParser.DASH)
                self.state = 146
                self.qnum()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 148
                self.typi()
                self.state = 149
                self.match(GlycanParser.NUM)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AddContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EQ(self):
            return self.getToken(GlycanParser.EQ, 0)

        def LBRACK(self):
            return self.getToken(GlycanParser.LBRACK, 0)

        def NUM(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.NUM)
            else:
                return self.getToken(GlycanParser.NUM, i)

        def RBRACK(self):
            return self.getToken(GlycanParser.RBRACK, 0)

        def ct(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.CtContext)
            else:
                return self.getTypedRuleContext(GlycanParser.CtContext,i)


        def COLON(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.COLON)
            else:
                return self.getToken(GlycanParser.COLON, i)

        def C(self):
            return self.getToken(GlycanParser.C, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_add

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAdd" ):
                listener.enterAdd(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAdd" ):
                listener.exitAdd(self)




    def add(self):

        localctx = GlycanParser.AddContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_add)
        self._la = 0 # Token type
        try:
            self.state = 181
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [31]:
                self.enterOuterAlt(localctx, 1)
                self.state = 153
                self.match(GlycanParser.EQ)
                self.state = 154
                self.match(GlycanParser.LBRACK)
                self.state = 156
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==27 or _la==30:
                    self.state = 155
                    self.ct()


                self.state = 158
                self.match(GlycanParser.NUM)
                self.state = 166
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==17:
                    self.state = 159
                    self.match(GlycanParser.COLON)
                    self.state = 161
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la==27 or _la==30:
                        self.state = 160
                        self.ct()


                    self.state = 163
                    self.match(GlycanParser.NUM)
                    self.state = 168
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 169
                self.match(GlycanParser.RBRACK)
                pass
            elif token in [27]:
                self.enterOuterAlt(localctx, 2)
                self.state = 170
                self.match(GlycanParser.C)
                self.state = 171
                self.match(GlycanParser.LBRACK)
                self.state = 172
                self.match(GlycanParser.NUM)
                self.state = 177
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==17:
                    self.state = 173
                    self.match(GlycanParser.COLON)
                    self.state = 174
                    self.match(GlycanParser.NUM)
                    self.state = 179
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 180
                self.match(GlycanParser.RBRACK)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class FgiContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COUNT(self):
            return self.getToken(GlycanParser.COUNT, 0)

        def bridge(self):
            return self.getTypedRuleContext(GlycanParser.BridgeContext,0)


        def FG(self):
            return self.getToken(GlycanParser.FG, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_fgi

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFgi" ):
                listener.enterFgi(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFgi" ):
                listener.exitFgi(self)




    def fgi(self):

        localctx = GlycanParser.FgiContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_fgi)
        try:
            self.state = 186
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [13]:
                self.enterOuterAlt(localctx, 1)
                self.state = 183
                self.match(GlycanParser.COUNT)
                pass
            elif token in [4, 5, 6, 7]:
                self.enterOuterAlt(localctx, 2)
                self.state = 184
                self.bridge()
                pass
            elif token in [8]:
                self.enterOuterAlt(localctx, 3)
                self.state = 185
                self.match(GlycanParser.FG)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CarbContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CARBON(self):
            return self.getToken(GlycanParser.CARBON, 0)

        def NUM(self):
            return self.getToken(GlycanParser.NUM, 0)

        def add(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.AddContext)
            else:
                return self.getTypedRuleContext(GlycanParser.AddContext,i)


        def AI(self):
            return self.getToken(GlycanParser.AI, 0)

        def A(self):
            return self.getToken(GlycanParser.A, 0)

        def I(self):
            return self.getToken(GlycanParser.I, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_carb

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCarb" ):
                listener.enterCarb(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCarb" ):
                listener.exitCarb(self)




    def carb(self):

        localctx = GlycanParser.CarbContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_carb)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 189
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 4395630592) != 0):
                self.state = 188
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 4395630592) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 191
            self.match(GlycanParser.CARBON)
            self.state = 192
            self.match(GlycanParser.NUM)
            self.state = 196
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==27 or _la==31:
                self.state = 193
                self.add()
                self.state = 198
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ModiContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NUM(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.NUM)
            else:
                return self.getToken(GlycanParser.NUM, i)

        def DASH(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.DASH)
            else:
                return self.getToken(GlycanParser.DASH, i)

        def ANHYDRO(self):
            return self.getToken(GlycanParser.ANHYDRO, 0)

        def COLON(self):
            return self.getToken(GlycanParser.COLON, 0)

        def bridge(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(GlycanParser.BridgeContext)
            else:
                return self.getTypedRuleContext(GlycanParser.BridgeContext,i)


        def fgi(self):
            return self.getTypedRuleContext(GlycanParser.FgiContext,0)


        def D(self):
            return self.getToken(GlycanParser.D, 0)

        def E(self):
            return self.getToken(GlycanParser.E, 0)

        def carb(self):
            return self.getTypedRuleContext(GlycanParser.CarbContext,0)


        def HEAD(self):
            return self.getToken(GlycanParser.HEAD, 0)

        def HEADD(self):
            return self.getToken(GlycanParser.HEADD, 0)

        def END(self):
            return self.getToken(GlycanParser.END, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_modi

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterModi" ):
                listener.enterModi(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitModi" ):
                listener.exitModi(self)




    def modi(self):

        localctx = GlycanParser.ModiContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_modi)
        self._la = 0 # Token type
        try:
            self.state = 242
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,24,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 201
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,18,self._ctx)
                if la_ == 1:
                    self.state = 199
                    self.match(GlycanParser.NUM)
                    self.state = 200
                    self.match(GlycanParser.COLON)


                self.state = 203
                self.match(GlycanParser.NUM)
                self.state = 204
                self.match(GlycanParser.DASH)
                self.state = 205
                self.match(GlycanParser.ANHYDRO)
                self.state = 206
                self.match(GlycanParser.DASH)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 207
                self.match(GlycanParser.NUM)
                self.state = 208
                self.match(GlycanParser.DASH)
                self.state = 209
                self.bridge()
                self.state = 210
                self.match(GlycanParser.DASH)
                self.state = 211
                self.fgi()
                self.state = 212
                self.match(GlycanParser.DASH)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 215
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==18:
                    self.state = 214
                    self.match(GlycanParser.DASH)


                self.state = 218
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==16:
                    self.state = 217
                    self.match(GlycanParser.NUM)


                self.state = 223
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,21,self._ctx)
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1:
                        self.state = 220
                        self.bridge() 
                    self.state = 225
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,21,self._ctx)

                self.state = 226
                self.fgi()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 227
                self.match(GlycanParser.NUM)
                self.state = 235
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [17, 28]:
                    self.state = 230
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la==17:
                        self.state = 228
                        self.match(GlycanParser.COLON)
                        self.state = 229
                        self.match(GlycanParser.NUM)


                    self.state = 232
                    self.match(GlycanParser.D)
                    pass
                elif token in [29]:
                    self.state = 233
                    self.match(GlycanParser.E)
                    pass
                elif token in [4, 25, 26, 32]:
                    self.state = 234
                    self.carb()
                    pass
                else:
                    raise NoViableAltException(self)

                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 237
                self.match(GlycanParser.HEAD)
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 238
                self.match(GlycanParser.HEADD)
                self.state = 239
                self.match(GlycanParser.DASH)
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 240
                self.match(GlycanParser.DASH)
                self.state = 241
                self.match(GlycanParser.END)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class QnumContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def QMARK(self):
            return self.getToken(GlycanParser.QMARK, 0)

        def NUM(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.NUM)
            else:
                return self.getToken(GlycanParser.NUM, i)

        def SLASH(self, i:int=None):
            if i is None:
                return self.getTokens(GlycanParser.SLASH)
            else:
                return self.getToken(GlycanParser.SLASH, i)

        def getRuleIndex(self):
            return GlycanParser.RULE_qnum

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQnum" ):
                listener.enterQnum(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQnum" ):
                listener.exitQnum(self)




    def qnum(self):

        localctx = GlycanParser.QnumContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_qnum)
        self._la = 0 # Token type
        try:
            self.state = 253
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [34]:
                self.enterOuterAlt(localctx, 1)
                self.state = 244
                self.match(GlycanParser.QMARK)
                pass
            elif token in [16]:
                self.enterOuterAlt(localctx, 2)
                self.state = 245
                self.match(GlycanParser.NUM)
                self.state = 250
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==33:
                    self.state = 246
                    self.match(GlycanParser.SLASH)
                    self.state = 247
                    self.match(GlycanParser.NUM)
                    self.state = 252
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class TypiContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TYPE(self):
            return self.getToken(GlycanParser.TYPE, 0)

        def QMARK(self):
            return self.getToken(GlycanParser.QMARK, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_typi

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTypi" ):
                listener.enterTypi(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTypi" ):
                listener.exitTypi(self)




    def typi(self):

        localctx = GlycanParser.TypiContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_typi)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 255
            _la = self._input.LA(1)
            if not(_la==14 or _la==34):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BridgeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CARBON(self):
            return self.getToken(GlycanParser.CARBON, 0)

        def NITROGEN(self):
            return self.getToken(GlycanParser.NITROGEN, 0)

        def OXYGEN(self):
            return self.getToken(GlycanParser.OXYGEN, 0)

        def PHOSPHOR(self):
            return self.getToken(GlycanParser.PHOSPHOR, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_bridge

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBridge" ):
                listener.enterBridge(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBridge" ):
                listener.exitBridge(self)




    def bridge(self):

        localctx = GlycanParser.BridgeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_bridge)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 257
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 240) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CtContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def C(self):
            return self.getToken(GlycanParser.C, 0)

        def T(self):
            return self.getToken(GlycanParser.T, 0)

        def getRuleIndex(self):
            return GlycanParser.RULE_ct

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCt" ):
                listener.enterCt(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCt" ):
                listener.exitCt(self)




    def ct(self):

        localctx = GlycanParser.CtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_ct)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 259
            _la = self._input.LA(1)
            if not(_la==27 or _la==30):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx





